


package org.hawk.module.script;


import org.hawk.data.perf.HawkPerfDataCollector;
import org.hawk.module.IModule;
import org.hawk.module.ModuleFactory;
import org.hawk.module.annotation.SubTask;
import org.hawk.module.script.enumeration.BuildModeEnum;
import org.hawk.module.task.SubTaskContainer;
import org.hawk.pattern.PatternMatcher;
import org.hawk.startup.HawkStartup;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import org.hawk.logger.HawkLogger;
import org.hawk.exception.*;
import org.hawk.module.script.enumeration.VarTypeEnum;
import org.hawk.module.script.type.Variable;
import org.hawk.shutdown.HawkShutdown;
import org.hawk.util.HawkUtil;


/**
 * <p>This parses the hawk script and runs the same.
 * This reads through the hawk file and parses the function definition and
 * hence creates the function script out of them.
 * The function script contains all the scripts written inside the function definition.
 *
 * <p>This is the place where all the global variables and aliases are stored.
 *
 * <p>This parses through target application's <tt>IModule</tt> implementations
 * to cache all the modules <tt>IModule</tt> and subtask <tt>SubTaskContainer</tt>
 * Note: The developer has to annotate the tasks with <tt>SubTask</tt> to register with
 * Hawk framework as subtask failing to which Hawk will ignore the same.
 *
 * 
 * @version 1.0 7 Apr, 2010
 * @author msahu
 * @see ScriptUsage
 * @see IModule
 * @see SubTaskContainer
 * @see SubTask
 */
public class ScriptInterpreter {

    private static final HawkLogger logger = HawkLogger.getLogger(ScriptInterpreter.class.getName());

    /**
     * Creates the singleton instance of <tt>ScriptInterpreter</tt>
     */
    private static final ScriptInterpreter scriptInterpreter = new ScriptInterpreter();

    /**
     * Private CTOR of <tt>ScriptInterpreter</tt>
     */
    private ScriptInterpreter(){

    }

    /**
     * SingleTon accessor method of <tt>ScriptInterpreter</tt>
     * @return
     */
    public static ScriptInterpreter getInstance(){
        return scriptInterpreter;
    }


    /**
     * A map containing key as <tt>IModule</tt>
     * and value as another map which contains
     * task name and <tt>SubTaskContainer</tt>
     * @see SubTaskContainer
     * @see IModule
     */
    private  Map<IModule,Map<String,SubTaskContainer>> cachedSubTasks =  new HashMap<IModule,Map<String,SubTaskContainer>>();
    
    /**
     * This stores the entire hawk script file's contents
     * in a map with key value as line no and line respectively.
     */
    private  Map<Integer,String> scriptMap = null;

    
    /**
     * A map containing key as function name
     * and value as <tt>MultiLineContainer</tt>
     * @see MultiLineContainer
     */
    private  Map<String,StructureDefnScript> structureDefnMap = null;

    /**
     * A map containing key as function name
     * and value as <tt>FunctionScript</tt>
     * @see FunctionScript
     */
    private  Map<String,FunctionScript> functionScriptMap = null;

    /**
     * Global alias map containing <tt>IModule</tt> and its
     * alias
     * @see AliasScript
     */
    private  Map<String ,String> aliasMap =new HashMap<String,String>();

    /**
     * Global map containing global var and its
     * value
     * @see VariableDeclScript
     */
    protected Map<Variable,IScript> globalVariableTable = new LinkedHashMap<Variable,IScript>();

    
    /**
     * A flag to check whether a hawk script can be interpreted.
     */
    private  boolean iterpretable = false;

    /**
     * A reference to main function which is the entry point
     * of hawk script execution.
     */
    private  FunctionScript mainFunction = null;

    public Map<Integer, String> getScriptMap() {
        return scriptMap;
    }

    public void setScriptMap(Map<Integer, String> scriptMap) {
        this.scriptMap = scriptMap;
    }


    /**
     * Getter for global variable
     * @param globalVar
     * @return getter for global variable
     */
    public IScript getGlobalValue(String globalVar){
        if(this.globalVariableTable == null && this.globalVariableTable.isEmpty()){
            return null;
        }
        Variable variable = new Variable(VarTypeEnum.VAR,null,globalVar);
        return this.globalVariableTable.get(variable);
    }

    /**
     * Setter for global variable
     * @param globalVar
     * @param globalValue
     */
    public void setGlobalValue(Variable globalVar,IScript globalValue){
        
        this.globalVariableTable.put(globalVar,globalValue);
    }

    /**
     * Removes the global variable from the global variable map.
     * @param globalVar the variable to be removed.
     */
    public void unsetGlobalValue(String globalVar){
        Variable variable = new Variable(VarTypeEnum.VAR,null,globalVar);
        this.globalVariableTable.remove(variable);
    }

    /**
     * Checks whether this global variable is declared.
     * @param globalVar
     * @return <tt>true</tt> on success <tt>false</tt> on failure
     */
    public boolean isGlobalVarDeclared(Variable globalVar){

        return this.globalVariableTable.containsKey(globalVar);
    }

    /**
     * This caches the subtasks implemented by the user using
     * <tt>SubTask</tt>.
     * @return a map containing key as <tt>IModule</tt>
     * and value as another map which contains
     * task name and <tt>SubTaskContainer</tt>
     */
    public  Map<IModule,Map<String,SubTaskContainer>> cacheSubTasks(){
        if(this.cachedSubTasks != null && !this.cachedSubTasks.isEmpty()){
            return this.cachedSubTasks;
        }

        Map<String,IModule> moduleMap = ModuleFactory.getModules();
        for(Entry<String,IModule> entry:moduleMap.entrySet()){

            IModule module = entry.getValue();

            Class moduleClazz = module.getClass();

            Method methods[] = moduleClazz.getDeclaredMethods();
            Map<String,SubTaskContainer> subTaskMap = new HashMap<String,SubTaskContainer>();
            for(Method method:methods){
                if(method.isAnnotationPresent(SubTask.class)){

                    SubTask subTaskAnnoation = (SubTask)method.getAnnotation(SubTask.class);


                    SubTaskContainer subTaskContainer = new SubTaskContainer();
                    subTaskContainer.setTaskName(subTaskAnnoation.name());
                    subTaskContainer.setSequence(subTaskAnnoation.sequence());
                    subTaskContainer.setTaskMethod(method);
                    subTaskContainer.setModule(module);
                    subTaskContainer.setIgnoreException(subTaskAnnoation.ignoreException());

                    subTaskMap.put(subTaskContainer.getTaskName(),subTaskContainer);
                }
            }

            this.cachedSubTasks.put(module, subTaskMap);

        }

        return this.cachedSubTasks;
    }

    /**
     * This returns the cached module->subtask map.
     * If it is not cached it invokes {@link ScriptInterpreter#cacheSubTasks}
     * @return a map containing key as <tt>IModule</tt>
     * and value as another map which contains
     * task name and <tt>SubTaskContainer</tt>
     */
    public  Map<IModule,Map<String,SubTaskContainer>> getSubTasks(){
        if(this.cachedSubTasks != null && !(this.cachedSubTasks.isEmpty())){
            return this.cachedSubTasks;
        }
        return cacheSubTasks();
    
    }

    /**
     * This returns {@link Set<SubTaskContainer>} the subtasks of a module sorted on the sequence number.
     * @param module for which sorted tasks are required.
     * @return returns {@link Set<SubTaskContainer>} the subtasks of a module sorted on the sequence number.
     * @throws org.hawk.exception.HawkException
     */
    public Set<SubTaskContainer> getSortedSubTasks(IModule module) throws HawkException{
        if(module == null){
            throw new HawkException("invalid module");
        }

        Map<IModule,Map<String,SubTaskContainer>> moduleTasks = this.getSubTasks();
        Map<String,SubTaskContainer> subTasks = moduleTasks.get(module);
        if(subTasks == null || subTasks.isEmpty()){
            throw new HawkException("no tasks for this module {"+ModuleFactory.getModuleName(module)+"}");
        }
        Set<SubTaskContainer> sortedTasks =  new TreeSet<SubTaskContainer>();

        for(Entry<String,SubTaskContainer> entry:subTasks.entrySet()){

            sortedTasks.add(entry.getValue());
        }

        return sortedTasks;
    }
   /**
     * This caches the structure definition
     * @return a map containing key as structure name
     * and value as <tt>StructureDefnScript</tt>
     * @throws org.hawk.exception.HawkException
     * @see StructureDefnScript
     */
    public  Map<String,StructureDefnScript> cacheStructureDefnMap() throws HawkException{
        this.structureDefnMap  = StructureDefnScript.parseStructures(this.scriptMap);
        for(Entry<String,StructureDefnScript>entry:this.structureDefnMap.entrySet()){
            StructureDefnScript.parseStructMembers(scriptMap, entry.getValue());
        }
        return this.structureDefnMap;
    }

    /**
     * This checks if structName is defined in the hawk script
     * @param structName
     * @return true if structName is defined otherwise false
     * @throws org.hawk.exception.HawkException
     */
    public boolean doesStructExist(String structName) throws HawkException{
        if(structName == null || structName.isEmpty()){
            throw new HawkException("Invalid struct name");
        }
        return this.getStructureDefnMap().containsKey(structName);
    }

    /**
     * This checks if a structName and it's member structMember are defined in the hawk script
     * @param structName
     * @param structMember
     * @return
     * @throws org.hawk.exception.HawkException
     */
    public boolean doesStructMemberExist(String structName,String structMember) throws HawkException{
        if(structName == null || structName.isEmpty()){
            throw new HawkException("Invalid struct name");
        }
        if(structMember == null || structMember.isEmpty()){
            throw new HawkException("Invalid struct member");
        }
        boolean status = false;
        if(this.doesStructExist(structName)){
            StructureDefnScript structDefnScript = this.getStructureDefnMap().get(structName);
            status = structDefnScript.doesMemberExist(structMember);
        }else{
            status = false;
        }
        return status;
    }

    /**
     * This returns the cached structure-><tt>StructureDefnScript</tt> map.
     * If it is not cached it invokes {@link ScriptInterpreter#cacheStructureDefnMap}
     * @return a map containing key as structure name
     * and value as <tt>StructureDefnScript</tt>
     * @throws org.hawk.exception.HawkException
     */
    public  Map<String,StructureDefnScript> getStructureDefnMap() throws HawkException{
        if(this.structureDefnMap != null && !(this.structureDefnMap.isEmpty())){
            return this.structureDefnMap;
        }
        return this.cacheStructureDefnMap();
    }

    
    /**
     * This clones input <tt>FunctionScript</tt>.
     * This is used in parallel execution of hawk functions
     * @see ExecParallelSingleLineScript
     * @param functionScript
     * @return
     */
    public FunctionScript cloneFunctionScript(FunctionScript functionScript){

        if(functionScript == null){
            return null;

        }
        FunctionScript clonedFunctionScript = functionScript.createFunctionTemplate();
        try {
            this.createFunctionScript(functionScript.mangle(), clonedFunctionScript);
        } catch (HawkException ex) {
           logger.info(ex.getMessage());
           return null;
        }

        return clonedFunctionScript;
    }

    /**
     * This caches the function script map
     * @return a map containing key as function name
     * and value as <tt>FunctionScript</tt>
     * @see FunctionScript
     * @throws org.hawk.exception.HawkException
     */
    public  Map<String,FunctionScript> cacheFunctionScriptMap() throws HawkException{

        this.functionScriptMap = FunctionScript.parseFunctions(this.scriptMap);

        for(Entry<String,FunctionScript> entry: this.functionScriptMap.entrySet()){
            if(this.createFunctionScript(entry.getKey(),entry.getValue())){
                this.functionScriptMap.put(entry.getKey(),entry.getValue());
            }

        }

        return this.functionScriptMap;
    }

    public boolean createFunctionScript(String functionName, FunctionScript functionScript) throws HawkException {

        ForLoopScript defaultForLoopScript = new ForLoopScript();
        defaultForLoopScript.setDefaultForLoop(true);
        functionScript.initializeParams();
        functionScript.setFunctionName(functionName);
        functionScript.setDefaultForLoopScript(defaultForLoopScript);
        defaultForLoopScript.setFunctionScript(functionScript);
        boolean parsed = false;
        try {
            parsed = MultiLineScript.parseMultiLines
                                        (
                                            this.scriptMap,
                                            functionScript.getMultiLineContainer().getStart() + 1,
                                            functionScript.getMultiLineContainer().getEnd(),
                                            defaultForLoopScript
                                        );
            if (parsed) {
                if (functionScript.isMainFunction()) {
                    this.iterpretable = true;
                    this.mainFunction = functionScript;
                }
                
            } else {
                throw new HawkException("parsing failure");
            }
        } catch (HawkException ex) {
            ex.printStackTrace();
            throw new HawkException("Compilation failed..."+ex.getMessage());
        }

        return true;
    }


    /**
     * This finds the <tt>FunctionScript</tt> for the input mangled function name
     * When invoked in threads other than main thread, this method clones the function
     * script and returns the same.
     * @param mangledFunctionName
     * @return returns <tt>FunctionScript</tt>
     * @throws HawkException 
     * @see FunctionScript
     */
    public FunctionScript findFunctionScript(String mangledFunctionName) throws HawkException{
        if(mangledFunctionName == null || mangledFunctionName.isEmpty()){
            throw new HawkException("Invalid function name");
        }

        FunctionScript functionScript = this.getFunctionScriptMap().get(mangledFunctionName);
        if(functionScript == null){
            throw new HawkException("Could not find function {"+mangledFunctionName.substring(0,mangledFunctionName.indexOf("_"))+"}");
        }

        if(!HawkUtil.isMainThread()){
            logger.info("Executing in thread {"+Thread.currentThread().getName()+"}");
            functionScript = this.cloneFunctionScript(functionScript);
        }

        return functionScript;
    }
    /**
     * This returns the cached function-><tt>FunctionScript</tt> map.
     * If it is not cached it invokes {@link ScriptInterpreter#cacheFunctionScriptMap}
     * @return a map containing key as function name
     * and value as <tt>FunctionScript</tt>
     * @throws org.hawk.exception.HawkException
     */
    public  Map<String,FunctionScript> getFunctionScriptMap() throws HawkException{
        if(this.functionScriptMap != null && !(this.functionScriptMap.isEmpty())){
            return this.functionScriptMap;
        }
        return cacheFunctionScriptMap();
    }

    /**
     * This caches all the global variables declared
     * @return
     * @throws org.hawk.exception.HawkException
     */
    public boolean cacheGlobalVariable() throws HawkException{
        boolean status = true;

        for(int i=0;i<this.scriptMap.size();i++){
            String scriptStr = this.scriptMap.get(i);
            if(!this.isInsideMultiLineScript(i)){

                 Map<Integer,String> lineCommentMatcherMap = null;
                 Map<Integer,String> lineGlobalVarMatcherMap = PatternMatcher.match
                                                                            (
                                                                                VariableDeclScript.getPattern(),
                                                                                scriptStr
                                                                            );
                 if(lineGlobalVarMatcherMap != null){
                     String globalVar = lineGlobalVarMatcherMap.get(1);
                     String globalVarExp = lineGlobalVarMatcherMap.get(2);
                     //FIX ME struct can be global vars....
                     Variable variable = new Variable(VarTypeEnum.VAR,null,globalVar);
                     VariableDeclScript globalVarScript = new VariableDeclScript();
                     globalVarScript.setVariable(variable);
                     globalVarScript.setVariableValue(globalVarScript.evaluateGlobalVariable(globalVarExp).getVariableValue());
                     globalVarScript.getVariable().setVariableValue(globalVarScript.getVariableValue().getVariableValue());
                     this.setGlobalValue(variable, globalVarScript);
                }else {
                
                      lineCommentMatcherMap = PatternMatcher.match
                                                            (
                                                                CommentScript.getPattern(),
                                                                scriptStr
                                                            );

                }

                if(
                        (lineGlobalVarMatcherMap== null && lineCommentMatcherMap==null)
                         &&
                        (scriptStr!=null && !scriptStr.trim().isEmpty())
                  ){
                    throw new HawkException ("Unrecognized characters {"+scriptStr+"} at line no {"+(++i)+"}");
                }
            }
            
        }
        return status;
    }

    private boolean isInsideMultiLineScript(int i){

        return this.isInsideFunctionScript(i) || this.isInsideStructureScript(i);
    }

    /**
     * This checks whether particular script is inside any function
     * definition block.
     * @param i line no at which script is present
     * @return <tt>true</tt>  if i is inside any function definition block
     *         <tt>false</tt> otherwise
     */
    private boolean isInsideFunctionScript(int i){

        boolean status = false;

        for(Entry<String,FunctionScript> entry:this.functionScriptMap.entrySet()){

           status = entry.getValue().isInside(i,true);
           if(status){
               break;
           }
        }


        return status;

    }

    /**
     * This checks whether particular script is inside any structure
     * definition block.
     * @param i line no at which script is present
     * @return <tt>true</tt>  if i is inside any structure definition block
     *         <tt>false</tt> otherwise
     */
    private boolean isInsideStructureScript(int i){

        boolean status = false;

        for(Entry<String,StructureDefnScript> entry:this.structureDefnMap.entrySet()){

           status = entry.getValue().isInside(i,true);
           if(status){
               break;
           }

        }


        return status;

    }

    /**
     * This checks file type extension of the input hawk script file
     * @param scriptFile hawk script file
     * @return <tt>true</tt>  if the hawk script file has .hawk extension
     *         
     * @throws org.hawk.exception.HawkException
     * if the hawk script file does not have .hawk extension
     */
    private  boolean validateHawkScript(String scriptFile) throws HawkException{

        if(!scriptFile.endsWith(".hawk")){

            throw new HawkException("Unrecognized hawk script file");
        }

        return true;
    }

    /**
     * This interprets the hawk script file.
     * @return
     * @throws org.hawk.exception.HawkException
     * if there is no main function defined in hawk script
     * or if the script is not interpretable.
     */
    public  boolean interpret() throws HawkException{
        boolean status = false;
        if(this.mainFunction == null || !this.iterpretable){
            throw new HawkException("Could not find main function");
        }

        this.mainFunction.executeDefaultForLoopScript(null);

        if(ScriptUsage.getInstance().getBuildMode() == BuildModeEnum.PERF){
            HawkPerfDataCollector.dump(true);
            status = HawkPerfDataCollector.processPerfData();
        }

        return status;

    }

    /**
     * This parses the hawk script file for syntax check.
     * @param scriptFile hawk script file
     * @return <tt>true</tt> if syntax check is successful
     *         <tt>false</tt> otherwise
     * @throws org.hawk.exception.HawkException
     *  if it is not valid hawk script file
     */
    public  boolean compile(String scriptFile) throws HawkException{


        if(this.validateHawkScript(scriptFile)){
            logger.info("Compiling {"+scriptFile+"}");
            this.readScript(scriptFile);
            this.cacheStructureDefnMap();
            this.cacheFunctionScriptMap();
            this.cacheGlobalVariable();
        }else{
            throw new HawkException("Unrecognized hawk script file");
        }
        

        return true;
    }

    /**
     * This reads the hawk script file and returns them as a map
     * with line no and script as key value pair respectively.
     * @param scriptFile hawk script file
     * @return a map containing line no and the script
     */
    private  Map<Integer,String> readScript(String scriptFile){
        this.scriptMap = new TreeMap<Integer,String>();

        BufferedReader bfr = null;
        
        try{

            bfr = new BufferedReader(new InputStreamReader(new FileInputStream(new File(scriptFile))));
            String line = null;
            int i =0;
            do{
                 line = bfr.readLine();
                 this.scriptMap.put(i, line);
                 i++;
            }while(line!=null);
            
        }catch(Exception ex){
            ex.printStackTrace();
            return null;
        }finally{
            if(bfr!=null){
                try {
                    bfr.close();
                } catch (IOException ex) {
                    ex.printStackTrace();
                    return null;
                }
            }
        }

        return this.scriptMap;
    }



    /**
     * Entry point of hawk script interpretation
     * @param args command line arguments as per java standards
     */
    public static void main(String args[]){
        
        Runtime.getRuntime().addShutdownHook(new Thread(new HawkShutdown()));
        
        boolean status = false;
        try{
            status = ScriptUsage.getInstance().parseArguments(args);
        }catch(Throwable ex){
            logger.info(ex.getMessage());
            return ;
        }
        if(!status){
            logger.info("Invalid option");
            ScriptUsage.getInstance().helpUser();
            return ;
        }
        BuildModeEnum buildMode = ScriptUsage.getInstance().getBuildMode();
        HawkStartup.getInstance().startUp();
        if(buildMode == BuildModeEnum.PERF){
            
        }else if(buildMode == BuildModeEnum.TRAINING){
            ScriptUsage.getInstance().showTasks();
            return ;
        }else if(buildMode == null ||buildMode == BuildModeEnum.HELP){
            ScriptUsage.getInstance().helpUser();
            return ;
        }


        ScriptInterpreter si = ScriptInterpreter.getInstance();
        boolean compilationStatus = false;
        try{
            String hawkScript = ScriptUsage.getInstance().getHawkScript();
            compilationStatus = si.compile(hawkScript);
        }catch(HawkException ex){

            logger.info(ex.getMessage());
            return ;
        }

        if(buildMode == BuildModeEnum.COMPILE){
            
            logger.info("Successfully compiled !!!");
            return;
        }else if(compilationStatus){
            try {
                 status = si.interpret();
                 
            } catch (Throwable th) {
                if(ScriptUsage.getInstance().getBuildMode() != BuildModeEnum.PERF){
                    th.printStackTrace();
                }else{
                    th.printStackTrace();
                    logger.severe("Failed during hawk script interpretation..."+th.getMessage());
                }
            }

            System.exit(status?0:1);

        }else{
            logger.info("Hawk script compilation failed....");
        }
        

    }

    
}




